@@ -26,12 +26,15 @@
 
 #include "display.h"
 
-#include "mpmath.h"
+#include "get.h"
+#include "mp.h"
 #include "functions.h"
 #include "ui.h"
 #include "ce_parser.h" // For ce_parse()
 #include "register.h"
 
+static const char *display_types[] = { "ENG", "FIX", "SCI", NULL };
+
 static GCDisplayState *
 get_state(GCDisplay *display)
 {
@@ -125,7 +128,7 @@ localize_expression(char *dest, const char *src, int dest_length, int *cursor)
         }
     }
 
-    if (!v->show_tsep) {
+    if (!v->display.show_tsep) {
         STRNCPY(dest, clean->str, dest_length - 1);
         g_string_free(clean, TRUE);
         return;
@@ -184,27 +187,18 @@ localize_expression(char *dest, const char *src, int dest_length, int *cursor)
 void
 display_clear(GCDisplay *display)
 {
+    v->error = 0;
     display_set_string(display, "", -1);
 }
 
 
-void
-display_reset(GCDisplay *display)
-{
-    v->error = 0;         /* Currently no display error. */
-    mp_set_from_integer(0, v->MPresult);   /* No previous result yet. */
-    mp_set_from_integer(0, v->MPdisp_val);         
-    mp_set_from_integer(0, v->MPlast_input);
-  
-    display_clear(display);
-}
-
 static const char *
 display_get_text(GCDisplay *display)
 {
     return get_state(display)->expression;
 }
 
+
 gboolean display_get_integer(GCDisplay *display, gint64 *value)
 {
     const char *text;
@@ -217,7 +211,7 @@ gboolean display_get_integer(GCDisplay *display, gint64 *value)
         text = "0";
     }
     else if (display_is_result(display)) {
-        make_number(buf, MAX_DISPLAY, display_get_answer(display), v->base, FALSE);
+        display_make_number(display, buf, MAX_DISPLAY, display_get_answer(display), v->base, FALSE);
         text = buf;
     }
     
@@ -227,6 +221,7 @@ gboolean display_get_integer(GCDisplay *display, gint64 *value)
     return TRUE;
 }
 
+
 gboolean display_get_unsigned_integer(GCDisplay *display, guint64 *value)
 {
     const char *text;
@@ -239,7 +234,7 @@ gboolean display_get_unsigned_integer(GCDisplay *display, guint64 *value)
         text = "0";
     }
     else if (display_is_result(display)) {
-        make_number(buf, MAX_DISPLAY, display_get_answer(display), v->base, FALSE);
+        display_make_number(display, buf, MAX_DISPLAY, display_get_answer(display), v->base, FALSE);
         text = buf;
     }
     
@@ -253,26 +248,67 @@ gboolean display_get_unsigned_integer(GCDisplay *display, guint64 *value)
     return TRUE;
 }
 
+
 int *display_get_answer(GCDisplay *display)
 {
     return get_state(display)->ans;
 }
 
+
 int
 display_get_cursor(GCDisplay *display)
 {
     return get_state(display)->cursor;
 }
 
+
 void
-display_set_number(GCDisplay *display, int *MPval)
+display_set_number(GCDisplay *display, const int *MPval)
 {
-    if (!v->error) {
-        make_number(display->display, MAX_DISPLAY, MPval, v->base, FALSE);
-        ui_set_display(display->display, -1);
-    }
+   char text[MAX_DISPLAY];
+   display_make_number(display, text, MAX_DISPLAY, MPval, v->base, FALSE);
+   display_set_string(display, text, -1);
 }
 
+
+static void
+display_refresh(GCDisplay *display)
+{
+    int i, MP_reg[MP_SIZE];
+    char localized[MAX_LOCALIZED], *str, reg[3], *t;
+    GCDisplayState *e;
+    char x[MAX_LOCALIZED], xx[MAX_LOCALIZED], ans[MAX_LOCALIZED];
+    int cursor = display_get_cursor(display);
+
+    e = get_state(display);
+    if (display_is_empty(display)) {
+        mp_set_from_integer(0, MP_reg);
+        display_make_number(display, x, MAX_LOCALIZED, MP_reg, v->base, FALSE);
+        str = x;
+    } else {           
+        str = strdup(e->expression);
+    }
+        
+    /* Substitute answer register */
+    display_make_number(display, ans, MAX_LOCALIZED, e->ans, v->base, TRUE);
+    localize_expression(localized, ans, MAX_LOCALIZED, &cursor);
+    str = str_replace(str, "Ans", localized);
+
+    /* Replace registers with values. */
+    for (i = 0; i < 10; i++) {
+        SNPRINTF(reg, 3, "R%d", i);
+        register_get(i, MP_reg);
+        display_make_number(display, xx, MAX_LOCALIZED, MP_reg, v->base, FALSE);
+        t = str_replace(str, reg, xx);
+        free(str);
+        str = t;
+    }
+    
+    ui_set_display(str, cursor);
+    free(str);
+}
+
+
 void
 display_set_string(GCDisplay *display, const char *value, int cursor)
 {
@@ -282,6 +318,8 @@ display_set_string(GCDisplay *display, const char *value, int cursor)
     free(e->expression);
     e->expression = strdup(value);
     e->cursor = cursor;
+
+    display_refresh(display);
 }
 
 void
@@ -291,6 +329,7 @@ display_set_cursor(GCDisplay *display, int cursor)
 
     e = get_state(display);
     e->cursor = cursor;
+    display_refresh(display);
 }
 
 void
@@ -299,13 +338,15 @@ display_set_error(GCDisplay *display, const char *message)
     ui_set_statusbar(message, "gtk-dialog-error");
 }
 
+
 static void
 copy_state(GCDisplayState *dst, GCDisplayState *src)
 {
-    MEMCPY(dst, src, sizeof(GCDisplayState));
+    memcpy(dst, src, sizeof(GCDisplayState));
     dst->expression = strdup(src->expression);
 }
 
+
 static void
 update_undo_redo_button_sensitivity(GCDisplay *display)
 {
@@ -323,6 +364,7 @@ update_undo_redo_button_sensitivity(GCDisplay *display)
     ui_set_undo_enabled(undo, redo);
 }
 
+
 void display_clear_stack(GCDisplay *display)
 {
     int i = display->h.begin;
@@ -337,6 +379,7 @@ void display_clear_stack(GCDisplay *display)
     update_undo_redo_button_sensitivity(display);
 }
 
+
 void display_push(GCDisplay *display)
 {
     int c;
@@ -365,6 +408,7 @@ void display_push(GCDisplay *display)
     update_undo_redo_button_sensitivity(display);
 }
 
+
 void display_pop(GCDisplay *display)
 {
     if (display->h.current != display->h.begin) {
@@ -378,6 +422,7 @@ void display_pop(GCDisplay *display)
     display_refresh(display);
 }
 
+
 void
 display_unpop(GCDisplay *display)
 {
@@ -388,8 +433,11 @@ display_unpop(GCDisplay *display)
         ui_set_statusbar(_("No redo steps"), "gtk-dialog-warning");
     }
     update_undo_redo_button_sensitivity(display);
+    get_state(display)->cursor = -1;
+    display_refresh(display);
 }
 
+
 gboolean
 display_is_undo_step(GCDisplay *display)
 {
@@ -397,10 +445,9 @@ display_is_undo_step(GCDisplay *display)
 }
 
 void
-display_insert(GCDisplay *display, const char *text)
+display_insert(GCDisplay *display, int cursor, const char *text)
 {
     char buf[MAX_DISPLAY], *currentText;
-    int cursor = display_get_cursor(display);
     
     if (cursor < 0) {
         SNPRINTF(buf, MAX_DISPLAY, "%s%s", display_get_text(display), text);
@@ -412,6 +459,15 @@ display_insert(GCDisplay *display, const char *text)
     display_set_string(display, buf, cursor);
 }
 
+void
+display_insert_number(GCDisplay *display, int cursor, const int value[MP_SIZE])
+{
+    char text[MAX_DISPLAY];
+    display_make_number(display, text, MAX_DISPLAY, value, v->base, FALSE);
+    display_insert(display, cursor, text);
+}
+
+
 void
 display_backspace(GCDisplay *display)
 {
@@ -424,7 +480,7 @@ display_backspace(GCDisplay *display)
     /* If cursor is at end of the line then delete the last character preserving accuracy */
     if (cursor < 0) {
         if (exp_has_postfix(e->expression, "Ans")) {
-            make_number(buf, MAX_DISPLAY, e->ans, v->base, FALSE);
+            display_make_number(display, buf, MAX_DISPLAY, e->ans, v->base, FALSE);
             t = str_replace(e->expression, "Ans", buf);
             free(e->expression);
             e->expression = t;
@@ -432,8 +488,8 @@ display_backspace(GCDisplay *display)
             for (i = 0; i < 10; i++) {
                 SNPRINTF(buf, MAX_DISPLAY, "R%d", i);
                 if (exp_has_postfix(e->expression, buf)) {
-                    do_rcl_reg(i, MP_reg);
-                    make_number(buf2, MAX_DISPLAY, MP_reg, v->base, FALSE);
+                    register_get(i, MP_reg);
+                    display_make_number(display, buf2, MAX_DISPLAY, MP_reg, v->base, FALSE);
                     /* Remove "Rx" postfix and replace with backspaced number */
                     SNPRINTF(buf, MAX_DISPLAY, "%.*s%s", strlen(e->expression) - 2, e->expression - 3, buf2);
                     display_set_string(display, buf, cursor - 1);
@@ -475,45 +531,6 @@ display_surround(GCDisplay *display, const char *prefix, const char *suffix)
     display_set_string(display, buffer, -1);
 }
 
-/* In arithmetic precedence mode this routine should be called to redraw 
- * the display.
- */
-void
-display_refresh(GCDisplay *display)
-{
-    int i, MP_reg[MP_SIZE];
-    char localized[MAX_LOCALIZED], *str, reg[3], *t;
-    GCDisplayState *e;
-    char x[MAX_LOCALIZED], xx[MAX_LOCALIZED], ans[MAX_LOCALIZED];
-    int cursor = display_get_cursor(display);
-
-    e = get_state(display);
-    if (display_is_empty(display)) {
-        mp_set_from_integer(0, MP_reg);
-        make_number(x, MAX_LOCALIZED, MP_reg, v->base, FALSE);
-        str = x;
-    } else {           
-        str = strdup(e->expression);
-    }
-        
-    /* Substitute answer register */
-    make_number(ans, MAX_LOCALIZED, e->ans, v->base, TRUE);
-    localize_expression(localized, ans, MAX_LOCALIZED, &cursor);
-    str = str_replace(str, "Ans", localized);
-
-    /* Replace registers with values. */
-    for (i = 0; i < 10; i++) {
-        SNPRINTF(reg, 3, "R%d", i);
-        do_rcl_reg(i, MP_reg);
-        make_number(xx, MAX_LOCALIZED, MP_reg, v->base, FALSE);
-        t = str_replace(str, reg, xx);
-        free(str);
-        str = t;
-    }
-    
-    ui_set_display(str, cursor);
-    free(str);
-}
 
 gboolean
 display_is_empty(GCDisplay *display)
@@ -540,16 +557,79 @@ display_is_usable_number(GCDisplay *display, int MPnum[MP_SIZE])
     }
 }
 
+
 void
 display_init(GCDisplay *display)
 {
     int i;
+   
+    memset(display, 0, sizeof(GCDisplay));
+   
+    display->base = 10;
+
+    if (get_boolean_resource(R_ZEROES, &i))
+        display->show_zeroes = i;
+    else
+        display->show_zeroes = FALSE;         
+
+    if (get_boolean_resource(R_TSEP, &i))
+        display->show_tsep = i;
+    else
+        display->show_tsep = FALSE;
+   
+    if (get_enumerated_resource(R_DISPLAY, display_types, &i))
+       display->format = (DisplayFormat) i;
+    else
+       display->format = FIX;
 
-    memset(&(display->h), 0, sizeof(GCDisplayHistory)); /* clear expression mode state history */
     for (i = 0; i < UNDO_HISTORY_LENGTH; i++)
         display->h.e[i].expression = strdup("");
 }
 
+
+void display_set_accuracy(GCDisplay *display, int accuracy)
+{
+    set_int_resource(R_ACCURACY, accuracy);
+    get_state(display)->cursor = -1;
+    display_refresh(display);   
+}
+
+
+void display_set_show_thousands_separator(GCDisplay *display, gboolean visible)
+{
+    display->show_tsep = visible;
+    set_boolean_resource(R_TSEP, visible);
+    display_set_cursor(display, -1);
+    display_refresh(display);
+}
+
+
+void display_set_show_trailing_zeroes(GCDisplay *display, gboolean visible)
+{
+    display->show_zeroes = visible;
+    set_boolean_resource(R_ZEROES, visible);
+    get_state(display)->cursor = -1;
+    display_refresh(display);
+}
+
+
+void display_set_base(GCDisplay *display, int base)
+{
+    display->base = base;
+    get_state(display)->cursor = -1;
+    display_refresh(display);
+}
+
+
+void display_set_format(GCDisplay *display, DisplayFormat type)
+{
+    v->display.format = type;
+    set_enumerated_resource(R_DISPLAY, display_types, (int) type);
+    get_state(display)->cursor = -1;
+    display_refresh(display);
+}
+
+
 int
 display_solve(GCDisplay *display, int *result)
 {
@@ -576,3 +656,137 @@ display_solve(GCDisplay *display, int *result)
     
     return errorCode;
 }
+
+
+/* Convert engineering or scientific number in the given base. */
+void
+make_eng_sci(GCDisplay *display, char *target, int target_len, const int *MPnumber, int base)
+{
+    static char digits[] = "0123456789ABCDEF";   
+    char fixed[MAX_DIGITS], *optr;
+    int MP1[MP_SIZE], MPatmp[MP_SIZE], MPval[MP_SIZE];
+    int MP1base[MP_SIZE], MP3base[MP_SIZE], MP10base[MP_SIZE];
+    int i, dval, len;
+    int MPmant[MP_SIZE];        /* Mantissa. */
+    int ddig;                   /* Number of digits in exponent. */
+    int eng = 0;                /* Set if this is an engineering number. */
+    int exp = 0;                /* Exponent */
+    
+    if (display->format == ENG) {
+        eng = 1;
+    }
+    optr = target;
+    mp_abs(MPnumber, MPval);
+    mp_set_from_integer(0, MP1);
+    if (mp_is_less_than(MPnumber, MP1)) {
+        *optr++ = '-';
+    }
+    mp_set_from_mp(MPval, MPmant);
+
+    mp_set_from_integer(basevals[base], MP1base);
+    mppwr(MP1base, 3, MP3base);
+
+    mppwr(MP1base, 10, MP10base);
+
+    mp_set_from_integer(1, MP1);
+    mpdiv(MP1, MP10base, MPatmp);
+
+    mp_set_from_integer(0, MP1);
+    if (!mp_is_equal(MPmant, MP1)) {
+        while (!eng && mp_is_greater_equal(MPmant, MP10base)) {
+            exp += 10;
+            mpmul(MPmant, MPatmp, MPmant);
+        }
+ 
+        while ((!eng &&  mp_is_greater_equal(MPmant, MP1base)) ||
+                (eng && (mp_is_greater_equal(MPmant, MP3base) || exp % 3 != 0))) {
+            exp += 1;
+            mpdiv(MPmant, MP1base, MPmant);
+        }
+ 
+        while (!eng && mp_is_less_than(MPmant, MPatmp)) {
+            exp -= 10;
+            mpmul(MPmant, MP10base, MPmant);
+        }
+ 
+        mp_set_from_integer(1, MP1);
+        while (mp_is_less_than(MPmant, MP1) || (eng && exp % 3 != 0)) {
+            exp -= 1;
+            mpmul(MPmant, MP1base, MPmant);
+        }
+    }
+ 
+    mp_cast_to_string(fixed, MAX_DIGITS, MPmant, basevals[base], v->accuracy);
+    len = strlen(fixed);
+    for (i = 0; i < len; i++) {
+        *optr++ = fixed[i];
+    }
+ 
+    *optr++ = 'e';
+ 
+    if (exp < 0) {
+        exp = -exp;
+        *optr++ = '-';
+    } else {
+        *optr++ = '+';
+    }
+ 
+    mp_set_from_string("0.5", 10, MP1);
+    mp_add_integer(MP1, exp, MPval);
+    mp_set_from_integer(1, MP1);
+    for (ddig = 0; mp_is_greater_equal(MPval, MP1); ddig++) {
+        mpdiv(MPval, MP1base, MPval);
+    }
+ 
+    if (ddig == 0) {
+        *optr++ = '0';
+    }
+ 
+    while (ddig-- > 0) {
+        mpmul(MPval, MP1base, MPval);
+        dval = mp_cast_to_int(MPval);
+        *optr++ = digits[dval];
+        dval = -dval;
+        mp_add_integer(MPval, dval, MPval);
+    }
+    *optr++    = '\0';
+}
+
+
+/* Convert MP number to character string in the given base. */
+void
+display_make_number(GCDisplay *display, char *target, int target_len, const int *MPnumber, int base, int ignoreError)
+{
+    static double max_fix[MAXBASES] = {
+       1.298074214e+33,    /* Binary. */
+       2.037035976e+90,    /* Octal. */
+       1.000000000e+100,   /* Decimal */
+       2.582249878e+120    /* Hexadecimal. */
+    };
+
+    double val;
+    
+    /*  NOTE: display_make_number can currently set v->error when converting to a double.
+     *        This is to provide the same look&feel as V3 even though gcalctool
+     *        now does internal arithmetic to "infinite" precision.
+     *
+     *  XXX:  Needs to be improved. Shouldn't need to convert to a double in
+     *        order to do these tests.
+     */
+
+    double number = mp_cast_to_double(MPnumber);
+
+    val = fabs(number);
+    if (v->error && !ignoreError) {
+        STRNCPY(target, _("Error"), target_len - 1);
+        return;
+    }
+    // FIXME: Do this based on the number of digits, not actual values
+    if ((display->format == ENG) ||
+        (display->format == SCI) ||
+        (display->format == FIX && val != 0.0 && (val > max_fix[base]))) {
+        make_eng_sci(display, target, target_len, MPnumber, base);
+    } else {
+        mp_cast_to_string(target, target_len, MPnumber, basevals[base], v->accuracy);
+    }
+}
