@@ -32,21 +32,20 @@
 #include "get.h"
 #include "register.h"
 #include "mp.h"
-#include "mpmath.h"
 #include "display.h"
 #include "ce_parser.h"
 #include "ui.h"
 
-enum button_flags {
+typedef enum {
     NUMBER       = (1 << 3),   /* Number button */
     FUNC         = (1 << 6),   /* Function */
     PREFIXOP     = (1 << 15),  /* Unary prefix operation */
-};
+} ButtonFlags;
 
 typedef struct {
     int id;
     char *symname;           /* Expression function name */
-    enum button_flags flags; /* Misc flags */
+    ButtonFlags flags; /* Misc flags */
 } Function;
 
 // FIXME: Sort this list
@@ -157,12 +156,10 @@ do_accuracy(int value)     /* Set display accuracy. */
 {
     v->accuracy = value;
     set_int_resource(R_ACCURACY, v->accuracy);
+    display_set_accuracy(&v->display, value);
     ui_set_accuracy(v->accuracy);
     ui_make_registers();
     clear_undo_history();
-
-    display_set_cursor(&v->display, -1);
-    display_refresh(&v->display);    
 }
 
 
@@ -189,25 +186,22 @@ do_shift(int count)     /* Perform bitwise shift on display value. */
 {
     int MPval[MP_SIZE];
 
-    if (display_is_usable_number(&v->display, MPval) || !is_integer(MPval)) {
+    if (display_is_usable_number(&v->display, MPval) || !mp_is_integer(MPval)) {
         /* Translators: This message is displayed in the status bar when a bit
            shift operation is performed and the display does not contain a number */
         ui_set_statusbar(_("No sane value to do bitwise shift"),
                          "gtk-dialog-error");
     }
     else {
-        calc_shift(MPval, display_get_answer(&v->display), count);
+        mp_shift(MPval, display_get_answer(&v->display), count);
         display_set_string(&v->display, "Ans", -1);
     }
-
-    display_set_cursor(&v->display, -1);
-    display_refresh(&v->display);
 }
 
 
 /* Change the current base setting. */
 static void
-do_base(enum base_type b)
+do_base(BaseType b)
 {
     int ret, MP[MP_SIZE];
 
@@ -221,13 +215,11 @@ do_base(enum base_type b)
         display_set_string(&v->display, "Ans", -1);
     }
     v->base = b;
-    set_resource(R_BASE, Rbstr[(int) v->base]);
+    set_enumerated_resource(R_BASE, Rbstr, (int) v->base);
+    display_set_base(&v->display, basevals[v->base]);
     ui_set_base(v->base);
     ui_make_registers();
     clear_undo_history();
-
-    display_set_cursor(&v->display, -1);
-    display_refresh(&v->display);
 }
 
 
@@ -246,23 +238,16 @@ do_exchange(int index)
         register_set(index, MPexpr);
         mp_set_from_mp(MPtemp, display_get_answer(&v->display));
         display_set_string(&v->display, "Ans", -1);
-        display_refresh(&v->display);
         ui_make_registers();
     }
-
-    display_set_cursor(&v->display, -1);
-    display_refresh(&v->display);
 }
 
 
 static void
-do_numtype(enum num_type n)   /* Set number display type. */
+do_numtype(DisplayFormat n)   /* Set number display type. */
 {
     int ret, MP[MP_SIZE];
 
-    v->dtype = n;
-    set_resource(R_DISPLAY, Rdstr[(int) v->dtype]);
-
     ret = display_is_usable_number(&v->display, MP);
     if (ret) {
         ui_set_statusbar(_("No sane value to convert"),
@@ -270,12 +255,11 @@ do_numtype(enum num_type n)   /* Set number display type. */
     } else {
         mp_set_from_mp(MP, display_get_answer(&v->display));
         display_set_string(&v->display, "Ans", -1);
-        ui_make_registers();
     }
     clear_undo_history();
-
-    display_set_cursor(&v->display, -1);
-    display_refresh(&v->display);
+   
+    display_set_format(&v->display, n);
+    ui_make_registers();
 }
 
 
@@ -284,10 +268,9 @@ do_sto(int index)
 {
     int temp[MP_SIZE];
     
-    if (display_is_usable_number(&v->display, temp)) {
+    if (display_is_usable_number(&v->display, temp))
         ui_set_statusbar(_("No sane value to store"),
                          "gtk-dialog-error");
-    }
     else
         register_set(index, temp);
 
@@ -335,7 +318,7 @@ do_expression(int function, int arg, int cursor)
         case FN_CLEAR_ENTRY:
             display_clear(&v->display);
             ui_set_error_state(FALSE);
-            MPstr_to_num("0", 10, ans);
+            mp_set_from_string("0", 10, ans);
             break;
 
         case FN_SHIFT:
@@ -352,7 +335,7 @@ do_expression(int function, int arg, int cursor)
         
         case FN_SET_TRIG_TYPE:
             v->ttype = arg;
-            set_resource(R_TRIG, Rtstr[arg]);
+            set_enumerated_resource(R_TRIG, Rtstr, arg);
             return;
 
         case FN_SET_NUMBERTYPE:
@@ -373,12 +356,11 @@ do_expression(int function, int arg, int cursor)
 
         case FN_RECALL:
             SNPRINTF(buf, MAXLINE, "R%d", arg);
-            display_insert(&v->display, buf);
+            display_insert(&v->display, display_get_cursor(&v->display), buf);
             break;
 
         case FN_CONSTANT:
-            make_number(buf, MAXLINE, constant_get_value(arg), v->base, FALSE);
-            display_insert(&v->display, buf);
+            display_insert_number(&v->display, display_get_cursor(&v->display), constant_get_value(arg));
             break;
 
         case FN_BACKSPACE:
@@ -404,13 +386,10 @@ do_expression(int function, int arg, int cursor)
 
                 bit_value ^= (1LL << (63 - arg));
     
-                /* FIXME: Convert to since we don't support setting MP numbers from 64 bit integers */
+                /* FIXME: Convert to string since we don't support setting MP numbers from 64 bit integers */
                 SNPRINTF(buf, MAX_DISPLAY, "%llu", bit_value);
-                MPstr_to_num(buf, 10, MP);
-
-                /* FIXME: Set as string as display_set_number doesn't store correctly */
-                make_number(buf, MAX_DISPLAY, MP, v->base, FALSE);
-                display_set_string(&v->display, buf, -1);
+                mp_set_from_string(buf, 10, MP);
+                display_set_number(&v->display, MP);
             }
             break;
 
@@ -473,7 +452,7 @@ do_expression(int function, int arg, int cursor)
             break;
 
         case FN_NUMERIC_POINT:
-            display_insert(&v->display, v->radix);
+            display_insert(&v->display, display_get_cursor(&v->display), v->radix);
             break;
 
         default:
@@ -484,16 +463,14 @@ do_expression(int function, int arg, int cursor)
             } else {
                 if (functions[function].flags & FUNC) {
                     SNPRINTF(buf, MAXLINE, "%s(", functions[function].symname);
-                    display_insert(&v->display, buf);
+                    display_insert(&v->display, display_get_cursor(&v->display), buf);
                 } else {
-                    display_insert(&v->display, functions[function].symname);
+                    display_insert(&v->display, display_get_cursor(&v->display), functions[function].symname);
                 }
             }
             break;
     }
 
-    display_refresh(&v->display);
-    
     enabled = display_get_unsigned_integer(&v->display, &bit_value);
     ui_set_bitfield(enabled, bit_value);
 }
